import pytest
import torch
from padl import transform, Batchify, batch, Identity
import padl.transforms as padl
from padl.util_transforms import IfTrain, IfEval, IfInfer, Try


@transform
def plus_one(x):
    return x + 1


@transform
def times_two(x):
    return x * 2


times_three = transform(lambda x: x * 3)


@transform
class TensorMult(torch.nn.Module):
    def __init__(self, factor):
        super().__init__()
        self.factor = factor

    def forward(self, x):
        return self.factor * x


@transform
class TensorDivide(torch.nn.Module):
    def __init__(self, div):
        super().__init__()
        self.div = div

    def forward(self, x):
        return x / self.div


class TestIfInMode:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = (
            plus_one
            >> IfTrain(times_two, plus_one)
            >> Batchify()
            >> times_three
        )
        request.cls.transform_2 = (
            plus_one
            >> IfEval(times_two)
            >> Batchify()
            >> times_three
        )
        request.cls.transform_3 = (
            plus_one
            >> IfInfer(times_two)
            >> Batchify()
            >> times_three
        )

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(1) == 9
        assert self.transform_2.infer_apply(1) == 6
        assert self.transform_3.infer_apply(1) == 12

    def test_eval_apply(self):
        assert list(self.transform_1.eval_apply([1, 2])) == [9, 12]
        assert list(self.transform_2.eval_apply([1, 2])) == [12, 18]
        assert list(self.transform_3.eval_apply([1, 2])) == [6, 9]

    def test_train_apply(self):
        assert list(self.transform_1.train_apply([1, 2])) == [12, 18]
        assert list(self.transform_2.train_apply([1, 2])) == [6, 9]
        assert list(self.transform_3.train_apply([1, 2])) == [6, 9]

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl', True)
        t1 = padl.load(tmp_path / 'test.padl')
        assert t1.infer_apply(1) == 9
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        t2 = padl.load(tmp_path / 'test.padl')
        assert t2.infer_apply(1) == 6
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        t3 = padl.load(tmp_path / 'test.padl')
        assert t3.infer_apply(1) == 12


class TestTry:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        tmp = transform(lambda x: x / 0)
        request.cls.try_transform_1 = (
            Try(tmp,
                times_two,
                (TypeError, ZeroDivisionError),
                plus_one,
                plus_one)
            >> Batchify()
            >> plus_one
        )
        request.cls.try_transform_2 = (
            plus_one
            >> transform(lambda x: torch.LongTensor([x]))
            >> Batchify()
            >> Try(TensorDivide('a'), TensorMult(2), TypeError, TensorMult(2), TensorMult(2))
        )

    def test_infer_apply(self):
        assert self.try_transform_1.infer_apply(4).item() == 9
        assert self.try_transform_2.infer_apply(3).item() == 8

    def test_train_apply(self):
        assert list(self.try_transform_1.train_apply([5, 8])) == [11, 17]
        assert list(self.try_transform_2.train_apply([2, 3, 4])) == [6, 8, 10]

    def test_eval_apply(self):
        assert list(self.try_transform_1.eval_apply([4, 9])) == [9, 19]
        assert list(self.try_transform_2.eval_apply([3, 7])) == [8, 16]

    def test_stages_1(self):
        tt = self.try_transform_1[0]
        assert isinstance(tt.pd_preprocess, Identity)
        assert tt.pd_forward is tt
        assert isinstance(tt.pd_postprocess, Identity)

    def test_no_multistage_try(self):
        with pytest.raises(AssertionError):
            t = Try(plus_one >> batch >> plus_one, Identity(), exceptions=Exception)
            t.pd_forward

    def test_no_multistage_catch(self):
        with pytest.raises(AssertionError):
            t = Try(Identity(), plus_one >> batch >> plus_one,
                    exceptions=Exception)
            t.pd_forward

    def test_no_multistage_else(self):
        with pytest.raises(AssertionError):
            t = Try(Identity(), Identity(), else_transform=plus_one >> batch >> plus_one,
                    exceptions=Exception)
            t.pd_forward

    def test_save_and_load(self, tmp_path):
        self.try_transform_1.pd_save(tmp_path / 'test.padl', force_overwrite=True)
        t1 = padl.load(tmp_path / 'test.padl')
        assert t1.infer_apply(4) == 9

        self.try_transform_2.pd_save(tmp_path / 'test.padl', force_overwrite=True)
        t2 = padl.load(tmp_path / 'test.padl')
        assert t2.infer_apply(3) == 8



"""Extra useful Transforms. """

from textwrap import indent
from collections import OrderedDict
from typing import Optional, Union, List, Tuple

from padl.transforms import ClassTransform, Identity, Transform, Mode, identity


class IfInMode(ClassTransform):
    """Perform *if_* if called in mode *target_mode*, else perform *else_*.

    Example:

        >>> from padl import transform
        >>> a = transform(lambda x: x + 10)
        >>> b = transform(lambda x: x * 99)
        >>> iim = IfInMode(a, 'infer', b)
        >>> iim.infer_apply(1)
        11
        >>> list(iim.eval_apply([1]))
        [99]

    :param if_: Transform to apply when the mode matches.
    :param target_mode: Mode (one of 'train', 'eval', 'infer').
    :param else_: Transform to apply when the mode doesn't match (defaults to identity transform).
    """

    def __init__(self, if_: Transform, target_mode: Mode, else_: Optional[Transform] = None):
        assert target_mode in ('train', 'eval', 'infer'), "Target mode can only be train, " \
                                                          "eval or infer"
        super().__init__(arguments=OrderedDict([('if_', if_),
                                                ('target_mode', target_mode),
                                                ('else_', else_)]))

        if else_ is None:
            else_ = Identity()

        self.if_ = if_
        self.else_ = else_
        self.target_mode = target_mode

    def _pd_longrepr(self, formatting=True, marker=None) -> str:
        if_repr = f'if pd_mode == "{self.target_mode}":\n\n' + indent(
            self.if_._pd_longrepr(marker=None), '   ')
        else_repr = f'\n\nelse:\n\n' + indent(self.else_._pd_longrepr(marker=None), '   ')
        if marker:
            return marker[1] + '\n\n' + if_repr + else_repr
        return if_repr + else_repr

    def __call__(self, args):
        assert Transform.pd_mode is not None, ('Mode is not set, use infer_apply, eval_apply '
                                               'or train_apply instead of calling the transform '
                                               'directly.')

        if Transform.pd_mode == self.target_mode:
            return self.if_._pd_unpack_args_and_call(args)
        return self.else_._pd_unpack_args_and_call(args)

    def _pd_splits(self, input_components=0):
        # pylint: disable=protected-access
        if_output_components, if_splits, if_has_batchify = \
            self.if_._pd_splits(input_components)
        else_output_components, else_splits, else_has_batchify = \
            self.else_._pd_splits(input_components)

        if_output_components_reduced = self._pd_merge_components(if_output_components)
        else_output_components_reduced = self._pd_merge_components(else_output_components)
        assert if_output_components_reduced == else_output_components_reduced, \
            'if and else transforms have incompatible output shapes'

        final_splits = tuple(
            IfInMode(
                if_=if_split,
                target_mode=self.target_mode,
                else_=else_split
            ) if if_split != else_split
            else if_split
            for if_split, else_split in zip(if_splits, else_splits)
        )

        return if_output_components_reduced, final_splits, if_has_batchify or else_has_batchify


class IfInfer(IfInMode):
    """Perform *if_* if called in "infer" mode, else perform *else_*.

    :param if_: Transform for the "infer" mode.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'infer', else_)


class IfEval(IfInMode):
    """Perform *if_* if called in "eval" mode, else perform *else_*.

    :param if_: Transform for the "eval" mode.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'eval', else_)


class IfTrain(IfInMode):
    """Perform *if_* if called in "train" mode, else perform *else_*.

    :param if_: Transform for the "train" mode.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'train', else_)


class Try(ClassTransform):
    """ Perform *transform*. If this fails with any exception from *exceptions*, perform
    *catch_transform*. If *transform* is completed successfully, *else_transform* is performed
    with the output of *transform*. Regardless of any error occurring on the other Transform
    (or not), *finally_transform* is carried out. No change of mode can happen inside any of these
    Transform.

    :param transform: Transform to try.
    :param catch_transform: Transform to fall back on.
    :param exceptions: Catch conditions.
    :param else_transform: Transform to carry on the `else` clause of the `try` statement.
    :param finally_transform: Transform to carry on the `finally` clause of the `try` statement.
    :param pd_name: The Transform's name.
    """

    def __init__(self,
                 transform: Transform,
                 catch_transform: Transform,
                 exceptions: Union[List, Tuple, Exception],
                 else_transform: Transform = identity,
                 finally_transform: Transform = identity,
                 pd_name: str = None):

        if not isinstance(exceptions, (tuple, list)):
            exceptions = (exceptions, )
        exceptions = tuple(exceptions)
        for exception in exceptions:
            assert issubclass(exception, Exception)
        super().__init__(pd_name=pd_name,
                         arguments=OrderedDict([('transform', transform),
                                                ('catch_transform', catch_transform),
                                                ('exceptions', exceptions),
                                                ('else_transform', else_transform),
                                                ('finally_transform', finally_transform)]))
        self.transform = transform
        self.catch_transform = catch_transform
        self.exceptions = exceptions
        self.else_transform = else_transform
        self.finally_transform = finally_transform

    def _pd_splits(self, input_components=0):
        try_output_components, _, _ = self.transform._pd_splits(input_components)
        catch_output_components, _, _ = self.catch_transform._pd_splits(input_components)
        else_output_components, _, _ = self.else_transform._pd_splits(input_components)

        input_components_reduced = self._pd_merge_components(input_components)
        try_output_components_reduced = self._pd_merge_components(try_output_components)
        catch_output_components_reduced = self._pd_merge_components(catch_output_components)
        else_output_components_reduced = self._pd_merge_components(else_output_components)
        components = [try_output_components_reduced, catch_output_components_reduced,
                      else_output_components_reduced]
        assert all(isinstance(component, int) for component in components) \
            and len(set(components)) == 1, \
            'Try Transform cannot contain transforms that have multiple stages.'

        final_splits = tuple(
            self if i == components[0]
            else identity
            for i in range(3)
        )

        return input_components_reduced, final_splits, False

    def _repr_exceptions(self):
        return '(' + ', '.join([exc.__name__ for exc in self.exceptions]) + ')'

    def _pd_longrepr(self, formatting=True, marker=None) -> str:
        tr_repr = f'try:\n\n' + indent(self.transform._pd_longrepr(marker=None), '   ')
        catch_tr_repr = f'\n\nexcept ' + self._repr_exceptions() + ':\n\n' + indent(
            self.catch_transform._pd_longrepr(marker=None), '   ')
        if isinstance(self.else_transform, Identity):
            else_tr_repr = ''
        else:
            else_tr_repr = f'\n\nelse:\n\n' + indent(
                self.else_transform._pd_longrepr(marker=None), '   ')
        if isinstance(self.finally_transform, Identity):
            finally_repr = ''
        else:
            finally_repr = f'\n\nfinally:\n\n' + indent(
                self.finally_transform._pd_longrepr(marker=None), '   ')
        if marker:
            return marker[1] + '\n\n' + tr_repr + catch_tr_repr + else_tr_repr + finally_repr
        return tr_repr + catch_tr_repr + else_tr_repr + finally_repr

    def __call__(self, args):
        try:
            output = self.transform._pd_unpack_args_and_call(args)
        except self.exceptions:
            output = self.catch_transform._pd_unpack_args_and_call(args)
        else:
            output = self.else_transform._pd_unpack_args_and_call(output)
        finally:
            self.finally_transform._pd_unpack_args_and_call(args)
        return output



"""Utilities. """

from padl.dumptools import inspector
from padl.print_utils import format_argument
from padl.transforms import AtomicTransform
from padl.transforms import _pd_trace


def _maketrans(attr, getitem=False):
    class T(AtomicTransform):
        """Dynamically generated transform for the "same" object.

        :param args: Arguments to pass to the input's method.
        :param kwargs: Keyword arguments to pass to the input's method.
        """

        def __init__(self, *args, **kwargs):
            self._args = args
            self._kwargs = kwargs
            caller_frameinfo = inspector.outer_caller_frameinfo(__name__)
            call_info = inspector.CallInfo(caller_frameinfo)
            if getitem:
                call = inspector.get_segment_from_frame(caller_frameinfo.frame, 'getitem')
            else:
                call = inspector.get_segment_from_frame(caller_frameinfo.frame, 'call')
            AtomicTransform.__init__(
                self,
                call=call,
                call_info=call_info,
            )

        def __call__(self, args):
            return getattr(args, attr)(*self._args, **self._kwargs)

        def _formatted_args(self) -> str:
            """Format the object's init arguments for printing. """
            args_list = [format_argument(val) for val in self._args]
            args_list += [f'{key}={format_argument(val)}' for key, val in self._kwargs.items()]
            return ', '.join(args_list)

        def _pd_longrepr(self, formatting=True, marker=None):
            out = self._pd_shortrepr()
            return out + marker[1] if marker else out

        def _pd_shortrepr(self, formatting=True, max_width=None):
            return f'{attr}({self._formatted_args()})'

        def _pd_tinyrepr(self, formatting=True, max_width=None):
            return self.pd_name or attr

    return T


class _Same:
    """Transform factory for capturing attributes/ get-items. """

    def __getitem__(self, item):
        return _maketrans('__getitem__', True)(item)

    def __getattr__(self, attr):
        return _maketrans(attr)


#: Transform factory for capturing attributes/ get-items.
same = _Same()


class _Debug:
    """Customized debugger for :class:`padl.transforms.Transform`s.

    When an exception on the execution of a :class:`padl.transforms.Transform` is produced and a
    :class:`_Debug` object is called, an interactive debugger at different levels in the
    :class:`padl.transforms.Transform` is gotten.

    At the top, the user interacts with the entire transform and its absolute input. One level
    down, it goes directly to the stage that got the Exception (either to
    :meth:`padl.transforms.Transform.pd_preprocess`, :meth:`padl.transforms.Transform.pd_forward`,
    or :meth:`padl.transforms.Transform.pd_postprocess`) and each level deeper moves recursively
    inside the element that failed until the :class:`padl.transforms.AtomicTransform` that got the
    Exception is reached.
    """
    def __init__(self):
        self.trans = None
        self.args = None
        self.default_msg = (
            'Defined commands are: \n'
            '   u(p): step up\n'
            '   d(own): step down\n'
            '   w(here am I?): show code position\n'
            '   i(nput): show input here\n'
            '   r(epeat): repeat here (will produce the same exception)\n'
            '   t(ransform): displays the current transform\n'
            '   h(elp): print help about the commands\n'
            '   q(uit): quit'
        )

    def __call__(self) -> None:
        """Call me for getting an interactive debugger in case of error.

        User can give following input and expect response
            u(p): step up
            d(own): step down
            w(here am I?): show code position
            i(nput): show input here
            r(epeat): repeat here (will produce the same exception)
            h(elp): print help about the commands
            q(uit): quit'
        """
        pos = len(_pd_trace) - 1
        print(self.default_msg + '\n' + _pd_trace[pos].transform_str)

        while True:
            try:
                x = input('> ')
            except IndexError:
                continue
            if x == 'd':
                pos, msg = self._down_step(pos, _pd_trace)
            elif x == 'u':
                pos, msg = self._up_step(pos, _pd_trace)
            elif x == 'q':
                self.args = _pd_trace[pos].args
                self.trans = _pd_trace[pos].transform
                break
            elif x == 'w':
                msg = _pd_trace[pos].code_position
            elif x == 'i':
                msg = _pd_trace[pos].args
            elif x == 'r':
                self.args = _pd_trace[pos].args
                self.trans = _pd_trace[pos].transform
                # This 0 is because the last element carries a problem: when adding the last
                # element to _pd_trace *Transform.pd_mode* has been already set up to None again.
                self.repeat(_pd_trace[0].pd_mode, pos)
            elif x == 'h' or x == 'help':
                msg = self.default_msg
            elif x == 't':
                msg = _pd_trace[pos].transform_str
            else:
                i = _pd_trace[pos].args
                try:
                    code = compile(x, '', 'single')
                    exec(code)
                except Exception as err:
                    print(err)

            if x in {'d', 'u', 'w', 'i', 'h', 'help', 't'}:
                print(f'\n{msg}\n')

    def repeat(self, mode: str, pos: int) -> None:
        """Repeat the execution from the current position *pos* (the same Exception will be
        produced).

        :param mode: mode ('train', 'eval', 'infer').
        :param pos: level of the :class:`Transform` we are inspecting.
        """
        assert mode in ('train', 'eval', 'infer'), 'Mode should be "train", "eval" or "infer'
        _pd_trace.clear()
        if pos == len(_pd_trace) - 1:
            self._repeat_entire(mode)
        else:
            self._repeat_on_stage(mode)

    @staticmethod
    def _down_step(pos, pd_trace):
        if pos > 0:
            pos -= 1
            return pos, pd_trace[pos].transform_str
        return pos, 'Reached the bottom.'

    @staticmethod
    def _up_step(pos, pd_trace):
        if pos < len(pd_trace) - 1:
            pos += 1
            return pos, pd_trace[pos].transform_str
        return pos, 'Reached top level.'

    def _repeat_entire(self, mode):
        if mode == 'train':
            list(self.trans.train_apply(self.args, batch_size=len(self.args), num_workers=0))
        elif mode == 'eval':
            list(self.trans.eval_apply(self.args, batch_size=len(self.args), num_workers=0))
        elif mode == 'infer':
            self.trans.infer_apply(self.args)
        raise ValueError('Mode is not set, it should be "train", "eval" or "infer')

    def _repeat_on_stage(self, mode):
        self.trans.pd_call_in_mode(self.args, mode)


pd_debug = _Debug()



from padl import transform


@transform
def function(x):
    return x


@transform
class Class:
    def __init__(self, x):
        self.x = x

    def __call__(self, y):
        return self.x + y


var = 100
obj = Class(var)

pipeline = obj >> obj + function >> function / function


def makefunction():
    @transform
    def z(x):
        return x
    return z

CONST = 123


def k(y):
    return y


@transform
def y(y):
    return CONST + k(y)


def makeclass():
    @transform
    class MyClassTransform:
        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def calculate(self, input_):
            return self.a * self.b + input_

        def __call__(self, input_):
            return self.calculate(input_) + y(input_) + self.c(input_)

    return MyClassTransform


def makeclasstransform(a, b, c):
    @transform
    class MyClassTransform:
        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def calculate(self, input_):
            return self.a * self.b + input_

        def __call__(self, input_):
            return self.calculate(input_) + y(input_) + self.c(input_)

    return MyClassTransform(a, b, c)


def makefunction_squared():
    return makefunction()


def makeclass_squared():
    return makeclasstransform(1, 2, 3)


def wrap_transform():
    return transform(k)


def makelambda():
    return transform(lambda x: x)



from padl import transform

LABELS = [
    "tench", "goldfish", "great_white_shark", "tiger_shark", "hammerhead", "electric_ray",
    "stingray", "cock", "hen", "ostrich", "brambling", "goldfinch", "house_finch", "junco",
    "indigo_bunting", "robin", "bulbul", "jay", "magpie", "chickadee", "water_ouzel", "kite",
    "bald_eagle", "vulture", "great_grey_owl", "European_fire_salamander", "common_newt",
    "eft", "spotted_salamander", "axolotl", "bullfrog", "tree_frog", "tailed_frog", "loggerhead",
    "leatherback_turtle", "mud_turtle", "terrapin", "box_turtle", "banded_gecko", "common_iguana",
    "American_chameleon", "whiptail", "agama", "frilled_lizard", "alligator_lizard",
    "Gila_monster", "green_lizard", "African_chameleon", "Komodo_dragon", "African_crocodile",
    "American_alligator", "triceratops", "thunder_snake", "ringneck_snake", "hognose_snake",
    "green_snake", "king_snake", "garter_snake", "water_snake", "vine_snake", "night_snake",
    "boa_constrictor", "rock_python", "Indian_cobra", "green_mamba", "sea_snake", "horned_viper",
    "diamondback", "sidewinder", "trilobite", "harvestman", "scorpion",
    "black_and_gold_garden_spider", "barn_spider", "garden_spider", "black_widow", "tarantula",
    "wolf_spider", "tick", "centipede", "black_grouse", "ptarmigan", "ruffed_grouse",
    "prairie_chicken", "peacock", "quail", "partridge", "African_grey", "macaw",
    "sulphur-crested_cockatoo", "lorikeet", "coucal", "bee_eater", "hornbill", "hummingbird",
    "jacamar", "toucan", "drake", "red-breasted_merganser", "goose", "black_swan", "tusker",
    "echidna", "platypus", "wallaby", "koala", "wombat", "jellyfish", "sea_anemone", "brain_coral",
    "flatworm", "nematode", "conch", "snail", "slug", "sea_slug", "chiton", "chambered_nautilus",
    "Dungeness_crab", "rock_crab", "fiddler_crab", "king_crab", "American_lobster",
    "spiny_lobster", "crayfish", "hermit_crab", "isopod", "white_stork", "black_stork",
    "spoonbill", "flamingo", "little_blue_heron", "American_egret", "bittern", "crane", "limpkin",
    "European_gallinule", "American_coot", "bustard", "ruddy_turnstone", "red-backed_sandpiper",
    "redshank", "dowitcher", "oystercatcher", "pelican", "king_penguin", "albatross", "grey_whale",
    "killer_whale", "dugong", "sea_lion", "Chihuahua", "Japanese_spaniel", "Maltese_dog",
    "Pekinese", "Shih-Tzu", "Blenheim_spaniel", "papillon", "toy_terrier", "Rhodesian_ridgeback",
    "Afghan_hound", "basset", "beagle", "bloodhound", "bluetick", "black-and-tan_coonhound",
    "Walker_hound", "English_foxhound", "redbone", "borzoi", "Irish_wolfhound",
    "Italian_greyhound", "whippet", "Ibizan_hound", "Norwegian_elkhound", "otterhound", "Saluki",
    "Scottish_deerhound", "Weimaraner", "Staffordshire_bullterrier",
    "American_Staffordshire_terrier", "Bedlington_terrier", "Border_terrier",
    "Kerry_blue_terrier", "Irish_terrier", "Norfolk_terrier", "Norwich_terrier",
    "Yorkshire_terrier", "wire-haired_fox_terrier", "Lakeland_terrier", "Sealyham_terrier",
    "Airedale", "cairn", "Australian_terrier", "Dandie_Dinmont", "Boston_bull",
    "miniature_schnauzer", "giant_schnauzer", "standard_schnauzer", "Scotch_terrier",
    "Tibetan_terrier", "silky_terrier", "soft-coated_wheaten_terrier",
    "West_Highland_white_terrier", "Lhasa", "flat-coated_retriever",
    "curly-coated_retriever", "golden_retriever", "Labrador_retriever",
    "Chesapeake_Bay_retriever", "German_short-haired_pointer", "vizsla", "English_setter",
    "Irish_setter", "Gordon_setter", "Brittany_spaniel", "clumber", "English_springer",
    "Welsh_springer_spaniel", "cocker_spaniel", "Sussex_spaniel", "Irish_water_spaniel",
    "kuvasz", "schipperke", "groenendael", "malinois", "briard", "kelpie", "komondor",
    "Old_English_sheepdog", "Shetland_sheepdog", "collie", "Border_collie", "Bouvier_des_Flandres",
    "Rottweiler", "German_shepherd", "Doberman", "miniature_pinscher",
    "Greater_Swiss_Mountain_dog", "Bernese_mountain_dog", "Appenzeller", "EntleBucher", "boxer",
    "bull_mastiff", "Tibetan_mastiff", "French_bulldog", "Great_Dane", "Saint_Bernard",
    "Eskimo_dog", "malamute", "Siberian_husky", "dalmatian", "affenpinscher", "basenji", "pug",
    "Leonberg", "Newfoundland", "Great_Pyrenees", "Samoyed", "Pomeranian", "chow", "keeshond",
    "Brabancon_griffon", "Pembroke", "Cardigan", "toy_poodle", "miniature_poodle",
    "standard_poodle", "Mexican_hairless", "timber_wolf", "white_wolf", "red_wolf", "coyote",
    "dingo", "dhole", "African_hunting_dog", "hyena", "red_fox", "kit_fox", "Arctic_fox",
    "grey_fox", "tabby", "tiger_cat", "Persian_cat", "Siamese_cat", "Egyptian_cat", "cougar",
    "lynx", "leopard", "snow_leopard", "jaguar", "lion", "tiger", "cheetah", "brown_bear",
    "American_black_bear", "ice_bear", "sloth_bear", "mongoose", "meerkat", "tiger_beetle",
    "ladybug", "ground_beetle", "long-horned_beetle", "leaf_beetle", "dung_beetle",
    "rhinoceros_beetle", "weevil", "fly", "bee", "ant", "grasshopper", "cricket",
    "walking_stick", "cockroach", "mantis", "cicada", "leafhopper", "lacewing",
    "dragonfly", "damselfly", "admiral", "ringlet", "monarch", "cabbage_butterfly",
    "sulphur_butterfly", "lycaenid", "starfish", "sea_urchin", "sea_cucumber", "wood_rabbit",
    "hare", "Angora", "hamster", "porcupine", "fox_squirrel", "marmot", "beaver", "guinea_pig",
    "sorrel", "zebra", "hog", "wild_boar", "warthog", "hippopotamus", "ox", "water_buffalo",
    "bison", "ram", "bighorn", "ibex", "hartebeest", "impala", "gazelle", "Arabian_camel",
    "llama", "weasel", "mink", "polecat", "black-footed_ferret", "otter", "skunk", "badger",
    "armadillo", "three-toed_sloth", "orangutan", "gorilla", "chimpanzee", "gibbon", "siamang",
    "guenon", "patas", "baboon", "macaque", "langur", "colobus", "proboscis_monkey", "marmoset",
    "capuchin", "howler_monkey", "titi", "spider_monkey", "squirrel_monkey", "Madagascar_cat",
    "indri", "Indian_elephant", "African_elephant", "lesser_panda", "giant_panda", "barracouta",
    "eel", "coho", "rock_beauty", "anemone_fish", "sturgeon", "gar", "lionfish", "puffer",
    "abacus", "abaya", "academic_gown", "accordion", "acoustic_guitar", "aircraft_carrier",
    "airliner", "airship", "altar", "ambulance", "amphibian", "analog_clock", "apiary", "apron",
    "ashcan", "assault_rifle", "backpack", "bakery", "balance_beam", "balloon", "ballpoint",
    "Band_Aid", "banjo", "bannister", "barbell", "barber_chair", "barbershop", "barn",
    "barometer", "barrel", "barrow", "baseball", "basketball", "bassinet", "bassoon",
    "bathing_cap", "bath_towel", "bathtub", "beach_wagon", "beacon", "beaker", "bearskin",
    "beer_bottle", "beer_glass", "bell_cote", "bib", "bicycle-built-for-two", "bikini",
    "binder", "binoculars", "birdhouse", "boathouse", "bobsled", "bolo_tie", "bonnet",
    "bookcase", "bookshop", "bottlecap", "bow", "bow_tie", "brass", "brassiere", "breakwater",
    "breastplate", "broom", "bucket", "buckle", "bulletproof_vest", "bullet_train", "butcher_shop",
    "cab", "caldron", "candle", "cannon", "canoe", "can_opener", "cardigan", "car_mirror",
    "carousel", "carpenter\'s_kit", "carton", "car_wheel", "cash_machine", "cassette",
    "cassette_player", "castle", "catamaran", "CD_player", "cello", "cellular_telephone",
    "chain", "chainlink_fence", "chain_mail", "chain_saw", "chest", "chiffonier", "chime",
    "china_cabinet", "Christmas_stocking", "church", "cinema", "cleaver", "cliff_dwelling",
    "cloak", "clog", "cocktail_shaker", "coffee_mug", "coffeepot", "coil", "combination_lock",
    "computer_keyboard", "confectionery", "container_ship", "convertible", "corkscrew", "cornet",
    "cowboy_boot", "cowboy_hat", "cradle", "crane", "crash_helmet", "crate", "crib", "Crock_Pot",
    "croquet_ball", "crutch", "cuirass", "dam", "desk", "desktop_computer", "dial_telephone",
    "diaper", "digital_clock", "digital_watch", "dining_table", "dishrag", "dishwasher",
    "disk_brake", "dock", "dogsled", "dome", "doormat", "drilling_platform", "drum", "drumstick",
    "dumbbell", "Dutch_oven", "electric_fan", "electric_guitar", "electric_locomotive",
    "entertainment_center", "envelope", "espresso_maker", "face_powder", "feather_boa", "file",
    "fireboat", "fire_engine", "fire_screen", "flagpole", "flute", "folding_chair",
    "football_helmet", "forklift", "fountain", "fountain_pen", "four-poster", "freight_car",
    "French_horn", "frying_pan", "fur_coat", "garbage_truck", "gasmask", "gas_pump", "goblet",
    "go-kart", "golf_ball", "golfcart", "gondola", "gong", "gown", "grand_piano", "greenhouse",
    "grille", "grocery_store", "guillotine", "hair_slide", "hair_spray", "half_track", "hammer",
    "hamper", "hand_blower", "hand-held_computer", "handkerchief", "hard_disc", "harmonica",
    "harp", "harvester", "hatchet", "holster", "home_theater", "honeycomb", "hook", "hoopskirt",
    "horizontal_bar", "horse_cart", "hourglass", "iPod", "iron", "jack-o\'-lantern", "jean",
    "jeep", "jersey", "jigsaw_puzzle", "jinrikisha", "joystick", "kimono", "knee_pad", "knot",
    "lab_coat", "ladle", "lampshade", "laptop", "lawn_mower", "lens_cap", "letter_opener",
    "library", "lifeboat", "lighter", "limousine", "liner", "lipstick", "Loafer", "lotion",
    "loudspeaker", "loupe", "lumbermill", "magnetic_compass", "mailbag", "mailbox", "maillot",
    "maillot", "manhole_cover", "maraca", "marimba", "mask", "matchstick", "maypole", "maze",
    "measuring_cup", "medicine_chest", "megalith", "microphone", "microwave", "military_uniform",
    "milk_can", "minibus", "miniskirt", "minivan", "missile", "mitten", "mixing_bowl",
    "mobile_home", "Model_T", "modem", "monastery", "monitor", "moped", "mortar", "mortarboard",
    "mosque", "mosquito_net", "motor_scooter", "mountain_bike", "mountain_tent", "mouse",
    "mousetrap", "moving_van", "muzzle", "nail", "neck_brace", "necklace", "nipple", "notebook",
    "obelisk", "oboe", "ocarina", "odometer", "oil_filter", "organ", "oscilloscope", "overskirt",
    "oxcart", "oxygen_mask", "packet", "paddle", "paddlewheel", "padlock", "paintbrush", "pajama",
    "palace", "panpipe", "paper_towel", "parachute", "parallel_bars", "park_bench",
    "parking_meter", "passenger_car", "patio", "pay-phone", "pedestal", "pencil_box",
    "pencil_sharpener", "perfume", "Petri_dish", "photocopier", "pick", "pickelhaube",
    "picket_fence", "pickup", "pier", "piggy_bank", "pill_bottle", "pillow", "ping-pong_ball",
    "pinwheel", "pirate", "pitcher", "plane", "planetarium", "plastic_bag", "plate_rack", "plow",
    "plunger", "Polaroid_camera", "pole", "police_van", "poncho", "pool_table", "pop_bottle",
    "pot", "potter\'s_wheel", "power_drill", "prayer_rug", "printer", "prison", "projectile",
    "projector", "puck", "punching_bag", "purse", "quill", "quilt", "racer", "racket", "radiator",
    "radio", "radio_telescope", "rain_barrel", "recreational_vehicle", "reel", "reflex_camera",
    "refrigerator", "remote_control", "restaurant", "revolver", "rifle", "rocking_chair",
    "rotisserie", "rubber_eraser", "rugby_ball", "rule", "running_shoe", "safe", "safety_pin",
    "saltshaker", "sandal", "sarong", "sax", "scabbard", "scale", "school_bus", "schooner",
    "scoreboard", "screen", "screw", "screwdriver", "seat_belt", "sewing_machine", "shield",
    "shoe_shop", "shoji", "shopping_basket", "shopping_cart", "shovel", "shower_cap",
    "shower_curtain", "ski", "ski_mask", "sleeping_bag", "slide_rule", "sliding_door",
    "slot", "snorkel", "snowmobile", "snowplow", "soap_dispenser", "soccer_ball", "sock",
    "solar_dish", "sombrero", "soup_bowl", "space_bar", "space_heater", "space_shuttle", "spatula",
    "speedboat", "spider_web", "spindle", "sports_car", "spotlight", "stage", "steam_locomotive",
    "steel_arch_bridge", "steel_drum", "stethoscope", "stole", "stone_wall", "stopwatch", "stove",
    "strainer", "streetcar", "stretcher", "studio_couch", "stupa", "submarine", "suit", "sundial",
    "sunglass", "sunglasses", "sunscreen", "suspension_bridge", "swab", "sweatshirt",
    "swimming_trunks", "swing", "switch", "syringe", "table_lamp", "tank", "tape_player", "teapot",
    "teddy", "television", "tennis_ball", "thatch", "theater_curtain", "thimble", "thresher",
    "throne", "tile_roof", "toaster", "tobacco_shop", "toilet_seat", "torch", "totem_pole",
    "tow_truck", "toyshop", "tractor", "trailer_truck", "tray", "trench_coat", "tricycle",
    "trimaran", "tripod", "triumphal_arch", "trolleybus", "trombone", "tub", "turnstile",
    "typewriter_keyboard", "umbrella", "unicycle", "upright", "vacuum", "vase", "vault",
    "velvet", "vending_machine", "vestment", "viaduct", "violin", "volleyball", "waffle_iron",
    "wall_clock", "wallet", "wardrobe", "warplane", "washbasin", "washer", "water_bottle",
    "water_jug", "water_tower", "whiskey_jug", "whistle", "wig", "window_screen", "window_shade",
    "Windsor_tie", "wine_bottle", "wing", "wok", "wooden_spoon", "wool", "worm_fence", "wreck",
    "yawl", "yurt", "web_site", "comic_book", "crossword_puzzle", "street_sign", "traffic_light",
    "book_jacket", "menu", "plate", "guacamole", "consomme", "hot_pot", "trifle", "ice_cream",
    "ice_lolly", "French_loaf", "bagel", "pretzel", "cheeseburger", "hotdog", "mashed_potato",
    "head_cabbage", "broccoli", "cauliflower", "zucchini", "spaghetti_squash", "acorn_squash",
    "butternut_squash", "cucumber", "artichoke", "bell_pepper", "cardoon", "mushroom",
    "Granny_Smith", "strawberry", "orange", "lemon", "fig", "pineapple", "banana", "jackfruit",
    "custard_apple", "pomegranate", "hay", "carbonara", "chocolate_sauce", "dough", "meat_loaf",
    "pizza", "potpie", "burrito", "red_wine", "espresso", "cup", "eggnog", "alp", "bubble",
    "cliff", "coral_reef", "geyser", "lakeside", "promontory", "sandbar", "seashore", "valley",
    "volcano", "ballplayer", "groom", "scuba_diver", "rapeseed", "daisy", "yellow_lady\'s_slipper",
    "corn", "acorn", "hip", "buckeye", "coral_fungus", "agaric", "gyromitra", "stinkhorn",
    "earthstar", "hen-of-the-woods", "bolete", "ear", "toilet_tissue"
]


@transform
class MyT:
    def __init__(self, x):
        pass


t = MyT(123)



import ast
from collections.abc import Iterable
import inspect
import json
from pathlib import Path
import sys
from types import ModuleType
from typing import Any, Callable, List, Optional

from padl.dumptools import inspector, sourceget, var2mod, symfinder
from padl.dumptools.symfinder import ScopedName
from padl.dumptools.var2mod import CodeNode, CodeGraph


SCOPE = symfinder.Scope.toplevel(sys.modules[__name__])


class Serializer:
    """Serializer base class.

    :param val: The value to serialize.
    :param save_function: The function to use for saving *val*.
    :param load_function: The function to use for loading *val*.
    :param file_suffix: If set, a string that will be appended to the path.
    :param module: The module the serializer functions are defined in. Optional, default is to
        use the calling module.
    """

    store: List = []
    i: int = 0

    def __init__(self, val: Any, save_function: Callable, load_function: callable,
                 file_suffix: Optional[str] = None, module: Optional[ModuleType] = None):
        self.index = Serializer.i
        Serializer.i += 1
        self.store.append(self)
        self.val = val
        self.save_function = save_function
        self.file_suffix = file_suffix
        if module is None:
            module = inspector.caller_module()
        self.scope = symfinder.Scope.toplevel(module)
        self.load_codegraph = \
            var2mod.CodeGraph.build(ScopedName(load_function.__name__,
                                               symfinder.Scope.toplevel(load_function.__module__)))
        self.load_name = load_function.__name__
        super().__init__()

    def save(self, path: Path):
        """Save the serializer's value to *path*.

        Returns a codegraph containing code needed to load the value.
        """
        if path is None:
            path = Path('?')
        if self.file_suffix is not None:
            path = Path(str(path) + f'/{self.i}{self.file_suffix}')
        filename = self.save_function(self.val, path)
        if filename is None:
            assert self.file_suffix is not None, ('if no file file_suffix is passed to *value*, '
                                                  'the *save*-function must return a filename')
            filename = path.name
        if isinstance(filename, (str, Path)):
            complete_path = f"pathlib.Path(__file__).parent / '{filename}'"
        elif isinstance(filename, Iterable):
            complete_path = ('[pathlib.Path(__file__).parent / filename for filename in ['
                             + ', '.join(f"'{fn}'"
                                         for fn in filename)
                             + ']]')
        else:
            raise ValueError('The save function must return a filename, a list of filenames or '
                             'nothing.')
        return CodeGraph(
            {**self.load_codegraph,
             ScopedName(self.varname, self.scope):
                 CodeNode(source=f'{self.varname} = {self.load_name}({complete_path})',
                          globals_={ScopedName(self.load_name, self.scope)},
                          scope=self.scope,
                          name=self.varname),
             ScopedName('pathlib', SCOPE):
                 CodeNode(source='import pathlib',
                          globals_=set(),
                          ast_node=ast.parse('import pathlib').body[0],
                          scope=self.scope,
                          name='pathlib')}
        )

    @property
    def varname(self):
        """The varname to store in the dumped code. """
        return f'PADL_VALUE_{self.index}'

    @classmethod
    def save_all(cls, codegraph, path):
        """Save all values. """
        for codenode in list(codegraph.values()):
            for serializer in cls.store:
                if serializer.varname in codenode.source:
                    loader_graph = serializer.save(path)
                    codegraph.update(loader_graph)


def save_json(val, path):
    """Saver for json. """
    with open(path, 'w') as f:
        json.dump(val, f)


def load_json(path):
    """Loader for json. """
    with open(path) as f:
        return json.load(f)


def json_serializer(val):
    """Create a json serializer for *val*. """
    return Serializer(val, save_json, load_json, '.json', sys.modules[__name__])


def _serialize(val, serializer=None):
    if serializer is not None:
        return Serializer(val, *serializer).varname
    if hasattr(val, '__len__') and len(val) > 10:
        return json_serializer(val).varname
    return repr(val)


def value(val, serializer=None):
    """Helper function that marks things in the code that should be stored by value. """
    caller_frameinfo = inspector.outer_caller_frameinfo(__name__)
    _call, locs = inspector.get_segment_from_frame(caller_frameinfo.frame, 'call', True)
    source = sourceget.get_source(caller_frameinfo.filename)
    sourceget.put_into_cache(caller_frameinfo.filename, sourceget.original(source),
                             _serialize(val, serializer=serializer), *locs)
    return val
